{
    "namespace": "suzieq",
    "name": "time",
    "type": "record",
    "fields": [
        {
            "name": "sqvers",
            "type": "string",
            "partition": 0,
            "default": "2.0",
            "suppress": true,
            "description": "Schema version, not selectable"
        },
        {
            "name": "ntpSync",
            "type": "string",
            "display": 2
        },
        {
            "name": "ntpEnabled",
            "type": "string"
        },
        {
            "name": "ntpServer",
            "type": "string",
            "display": 3
        },
        {
            "name": "stratum",
            "type": "long",
            "display": 4
        },
        {
            "name": "offset",
            "type": "string",
            "display": 5
        },
        {
            "name": "timezone",
            "type": "string"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
            "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
            "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ]
}
