{
    "namespace": "suzieq",
    "name": "device",
    "type": "record",
    "fields": [
        {
            "name": "sqvers",
            "type": "string",
            "partition": 0,
            "default": "3.0",
            "suppress": true,
            "description": "Schema version, not selectable"
        },
        {
            "name": "architecture",
            "type": "string",
            "display": 5
        },
        {
            "name": "memory",
            "type": "int"
        },
        {
            "name": "reason",
            "type": "string"
        },
        {
            "name": "status",
            "type": "string",
            "display": 6,
            "description": "Status of device: alive/dead/neverpoll"
        },
        {
            "name": "address",
            "type": "string",
            "display": 7,
            "description": "The IP address/hostname used for polling"
        },
        {
            "name": "model",
            "type": "string",
            "display": 2
        },
        {
            "name": "vendor",
            "type": "string",
            "display": 4
        },
        {
            "name": "version",
            "type": "string",
            "display": 3
        },
        {
            "name": "bootupTimestamp",
            "type": "double",
            "display": 9
        },
        {
            "name": "uptime",
            "type": "double",
            "depends": "bootupTimestamp timestamp"
        },
        {
            "name": "os",
            "type": "string"
        },
        {
            "name": "serialNumber",
            "type": "string"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
            "partition": 2,
            "description": "Hostname associated with this record"
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
            "partition": 1,
            "description": "Namespace associated with this record"
        },
        {
            "name": "timestamp",
            "type": "timestamp",
            "display": 10
        },
        {
            "name": "active",
            "type": "boolean",
            "description": "If this entry is active or deleted"
        }
    ]
}
