{
    "namespace": "suzieq",
    "name": "arpnd",
    "type": "record",
    "fields": [
        {
            "name": "sqvers",
            "type": "string",
            "partition": 0,
            "default": "1.0",
            "suppress": true,
            "description": "Schema version, not selectable"
        },
        {
            "name": "state",
            "type": "string",
            "display": 5,
            "description": "State of entry: STALE, REACHABLE etc."
        },
        {
            "name": "ipAddress",
            "type": "string",
            "key": 2,
            "display": 2,
            "description": "IP Address of entry"
        },
        {
            "name": "oif",
            "type": "string",
            "key": 3,
            "display": 3,
            "description": "Name of outgoing interface"
        },
        {
            "name": "remote",
            "type": "boolean",
            "display": 6,
            "description": "True if entry is learnt via EVPN"
        },
        {
            "name": "macaddr",
            "type": "string",
            "display": 4,
            "description": "MAC address of entry"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
            "partition": 2,
            "description": "hostname associated with this record"
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
            "partition": 1,
            "description": "namespace associated with this record"
        },
        {
            "name": "timestamp",
            "type": "timestamp",
            "display": 7,
            "description": "Unix epach When this record was created, in ms"
        },
        {
            "name": "active",
            "type": "boolean",
            "description": "If this entry is active or deleted"
        }
    ]
}
